Experience = LibStub("AceAddon-3.0"):NewAddon("Experience", "AceEvent-3.0")
local L = LibStub("AceLocale-3.0"):GetLocale("Experience")
local LDB = LibStub:GetLibrary("LibDataBroker-1.1")
local ACD3 = LibStub("AceConfigDialog-3.0")
local mod = Experience

local weightLookup, dungeonKeys = {}, {}
local profile, db, inspectUnit, currentUnit, configFrame
local cache = {}

local tonumber = _G.tonumber
local pairs, ipairs = _G.pairs, _G.ipairs
local tinsert, type, math = _G.tinsert, _G.type, _G.math

local tooltip = CreateFrame("GameTooltip", "ExperienceTooltip", UIParent, "GameTooltipTemplate")
tinsert(UISpecialFrames,"ExperienceTooltip")
tooltip:EnableMouse(true)
tooltip:SetMovable(true)
tooltip:RegisterForDrag("LeftButton")
tooltip:SetPoint("LEFT")
tooltip:SetOwner(UIParent, "ANCHOR_PRESERVE")
tooltip:SetScript("OnShow", nil)
tooltip:SetScript("OnHide", nil)
tooltip.close = CreateFrame("Button", nil, tooltip, "UIPanelCloseButton")
tooltip.close:SetPoint("TOPRIGHT")
tooltip.close:SetScript("OnClick", function(self)
	self:GetParent():Hide()
end)
tooltip:SetScript("OnDragStart", function(self)
	self:StartMoving()
end)
tooltip:SetScript("OnDragStop", function(self)
	self:StopMovingOrSizing()
	ValidateFramePosition(self)
end)
tooltip:SetScript("OnLoad", function(self)
	GameTooltip_OnLoad(self)
	self:SetPadding(16)
end)

local defaults = {
	profile = {
		experiencedKills = 3,
		excludeDungeons = {},
		hideEmpty = true
	}
}

local options = {
	type = "group",
	args = {
		threshold = {
			name = L["Kill Threshold"],
			desc = L["Players with this many kills per boss will earn full credit towards experience in that instance."],
			type = "range",
			min = 1,
			max = 10,
			step = 1,
			bigStep = 1,
			get = function() return profile.experiencedKills end,
			set = function(info, v)
				profile.experiencedKills = v
			end
		},
		hideEmpty = {
			name = L["Hide empty"],
			desc = L["Hide entries that players have no experience in."],
			type = "toggle",
			get = function() return profile.hideEmpty end,
			set = function(info, v)
				profile.hideEmpty = v
			end,
			order = 150,
		},
		showDetail = {
			name = L["Show detail"],
			desc = L["Show detailed points for each dungeon"],
			type = "toggle",
			get = function() return profile.showDetail end,
			set = function(info, v)
				profile.showDetail = v
			end,
			order = 160,
		},
		dungeons = {
			type = "group",
			name = L["Dungeons"],
			desc = L["Dungeons to consider when calculating experience"],
			args = {}			
		}
	}
}

local ldbObj = LDB:NewDataObject("Experience", {
	type = "data source",
	icon = [[Interface\ICONS\Achievement_PVP_A_14]],
	label = "Experience",
	text  = "|cff00ff00Enabled|r",
	
	OnClick = function(self, btn)
		if btn == "RightButton" then
			InterfaceOptionsFrame_OpenToCategory(configFrame)
		else
			mod:Toggle()
		end
	end
})

local metas = {
	-- [L["5-Man Dungeons"]] = { 1242, 1231, 1232, 1233, 1234, 1235, 1236, 1237, 1238, 1239, 1240, 1241 },
	[L["5-Man Heroics"]] = { 1504, 1505, 1506, 1507, 1508, 1509, 1510, 1511, 1512, 1513, 1514, 1515 },
}

local dungeons = {
	-- [L["5-Man Dungeons"]] = {
		-- [L["Utgarde Keep"]] 				= { [1242] = 5 },
		-- [L["The Nexus"]] 					= { [1231] = 5 },
		-- [L["Azjol-Nerub"]] 					= { [1232] = 5 },
		-- [L["Ahn'kahet: The Old Kingdom"]] 	= { [1233] = 5 },
		-- [L["Drak'Tharon Keep"]] 			= { [1234] = 5 },
		-- [L["The Violet Hold"]] 				= { [1235] = 5 },
		-- [L["Gundrak"]] 						= { [1236] = 5 },
		-- [L["Halls of Stone"]] 				= { [1237] = 5 },
		-- [L["Halls of Lightning"]] 			= { [1238] = 5 },
		-- [L["The Oculus"]] 					= { [1239] = 5 },
		-- [L["Utgarde Pinnacle"]] 			= { [1240] = 5 },
		-- [L["Caverns of Time: Stratholme"]] 	= { [1241] = 5 }
	-- },
	[L["5-Man Heroics"]] = {
		[L["Utgarde Keep"]] 				= { [1504] = 5 },
		[L["The Nexus"]] 					= { [1505] = 5 },
		[L["Azjol-Nerub"]] 					= { [1506] = 5 },
		[L["Ahn'kahet: The Old Kingdom"]] 	= { [1507] = 5 },
		[L["Drak'Tharon Keep"]] 			= { [1508] = 5 },
		[L["The Violet Hold"]] 				= { [1509] = 5 },
		[L["Gundrak"]] 						= { [1510] = 5 },
		[L["Halls of Stone"]] 				= { [1511] = 5 },
		[L["Halls of Lightning"]] 			= { [1512] = 5 },
		[L["The Oculus"]] 					= { [1513] = 5 },
		[L["Utgarde Pinnacle"]] 			= { [1514] = 5 },
		[L["Caverns of Time: Stratholme"]] 	= { [1515] = 5 }
	},
	[L["Naxxramas (10)"]] = {
		[1361] = 2,		-- <statistic id="1368" name="Anub'Rekhan kills (Naxxramas 25 player)" quantity="15"/>
		[1362] = 2,		-- <statistic id="1362" name="Grand Widow Faerlina kills (Naxxramas 10 player)" quantity="8"/>
		[1363] = 5,		-- <statistic id="1363" name="Maexxna kills (Naxxramas 10 player)" quantity="8"/>
		[1364] = 2,		-- <statistic id="1364" name="Patchwerk kills (Naxxramas 10 player)" quantity="9"/>
		[1371] = 2,		-- <statistic id="1371" name="Grobbulus kills (Naxxramas 10 player)" quantity="8"/>
		[1372] = 2,		-- <statistic id="1372" name="Gluth kills (Naxxramas 10 player)" quantity="8"/>
		[1373] = 5,		-- <statistic id="1373" name="Thaddius kills (Naxxramas 10 player)" quantity="8"/>
		[1365] = 2,		-- <statistic id="1365" name="Noth the Plaguebringer kills (Naxxramas 10 player)" quantity="8"/>
		[1369] = 2,		-- <statistic id="1369" name="Heigan the Unclean kills (Naxxramas 10 player)" quantity="8"/>
		[1370] = 5,		-- <statistic id="1370" name="Loatheb kills (Naxxramas 10 player)" quantity="8"/>
		[1374] = 2,		-- <statistic id="1374" name="Instructor Razuvious kills (Naxxramas 10 player)" quantity="9"/>
		[1366] = 2,		-- <statistic id="1366" name="Gothik the Harvester kills (Naxxramas 10 player)" quantity="9"/>
		[1375] = 5,		-- <statistic id="1375" name="Four Horsemen kills (Naxxramas 10 player)" quantity="9"/>
		[1376] = 10,	-- <statistic id="1376" name="Sapphiron kills (Naxxramas 10 player)" quantity="8"/>
		[1377] = 10,	-- <statistic id="1377" name="Kel'Thuzad kills (Naxxramas 10 player)" quantity="8"/>
		meta = true
	},
	
	[L["Naxxramas (25)"]] = {
		[1368] = 2,		-- <statistic id="1368" name="Anub'Rekhan kills (Naxxramas 25 player)" quantity="15"/>
		[1380] = 2,		-- <statistic id="1362" name="Grand Widow Faerlina kills (Naxxramas 10 player)" quantity="8"/>
		[1386] = 5,		-- <statistic id="1363" name="Maexxna kills (Naxxramas 10 player)" quantity="8"/>
		[1367] = 2,		-- <statistic id="1364" name="Patchwerk kills (Naxxramas 10 player)" quantity="9"/>
		[1381] = 2,		-- <statistic id="1371" name="Grobbulus kills (Naxxramas 10 player)" quantity="8"/>
		[1378] = 2,		-- <statistic id="1372" name="Gluth kills (Naxxramas 10 player)" quantity="8"/>
		[1388] = 5,		-- <statistic id="1373" name="Thaddius kills (Naxxramas 10 player)" quantity="8"/>
		[1387] = 2,		-- <statistic id="1365" name="Noth the Plaguebringer kills (Naxxramas 10 player)" quantity="8"/>
		[1382] = 2,		-- <statistic id="1369" name="Heigan the Unclean kills (Naxxramas 10 player)" quantity="8"/>
		[1385] = 5,		-- <statistic id="1370" name="Loatheb kills (Naxxramas 10 player)" quantity="8"/>
		[1384] = 2,		-- <statistic id="1374" name="Instructor Razuvious kills (Naxxramas 10 player)" quantity="9"/>
		[1379] = 2,		-- <statistic id="1366" name="Gothik the Harvester kills (Naxxramas 10 player)" quantity="9"/>
		[1383] = 5,		-- <statistic id="1375" name="Four Horsemen kills (Naxxramas 10 player)" quantity="9"/>
		[1389] = 10,	-- <statistic id="1376" name="Sapphiron kills (Naxxramas 10 player)" quantity="8"/>
		[1390] = 10,	-- <statistic id="1377" name="Kel'Thuzad kills (Naxxramas 10 player)" quantity="8"/>
		meta = true
	},
	
	[L["Sartharion (10)"]] = { [1392] = 5, meta = true},
	[L["Malygos (10)"]] = { [1391] = 5, meta = true },
	[L["Vault of Archavon (10)"]] = {
		[1753] = 5,		-- Archavon
		[2870] = 5,		-- Emalon
		[4074] = 5,		-- Koralon
		meta = true
	},
	
	[L["Sartharion (25)"]] = { [1393] = 5, meta = true},
	[L["Malygos (25)"]] = { [1394] = 5, meta = true },
	[L["Vault of Archavon (25)"]] = {
		[1754] = 5,		-- Archavon
		[3236] = 5,		-- Emalon
		[4075] = 5,		-- Koralon
		meta = true
	},
	
	[L["Ulduar (10)"]] = {
		[2856] = 1,		-- <statistic id="2856" name="Flame Leviathan kills (Ulduar 10 player)" quantity="20"/>
		[2857] = 5,		-- <statistic id="2857" name="Razorscale kills (Ulduar 10 player)" quantity="6"/>
		[2858] = 5,		-- <statistic id="2858" name="Ignis the Furnace Master kills (Ulduar 10 player)" quantity="3"/>
		[2859] = 5,		-- <statistic id="2859" name="XT-002 Deconstructor kills (Ulduar 10 player)" quantity="6"/>
		[2860] = 10,	-- <statistic id="2860" name="Assembly of Iron kills (Ulduar 10 player)" quantity="12"/>
		[2861] = 5,		-- <statistic id="2861" name="Kologarn kills (Ulduar 10 player)" quantity="18"/>
		[2868] = 5,		-- <statistic id="2868" name="Auriaya kills (Ulduar 10 player)" quantity="19"/>
		[2862] = 10,	-- <statistic id="2862" name="Hodir victories (Ulduar 10 player)" quantity="13"/>
		[2863] = 10,	-- <statistic id="2863" name="Thorim victories (Ulduar 10 player)" quantity="21"/>
		[2864] = 10,	-- <statistic id="2864" name="Freya victories (Ulduar 10 player)" quantity="15"/>
		[2865] = 10,	-- <statistic id="2865" name="Mimiron victories (Ulduar 10 player)" quantity="9"/>
		[2866] = 15,	-- <statistic id="2866" name="General Vezax kills (Ulduar 10 player)" quantity="9"/>
		[2869] = 15,	-- <statistic id="2869" name="Yogg-Saron kills (Ulduar 10 player)" quantity="8"/>
		[2867] = 10, 	-- <statistic id="2867" name="Algalon the Observer kills (Ulduar 10 player)" quantity="6"/>
		baseline = -10,
		meta = true
	},
	[L["Ulduar (25)"]] = {
		[2872] = 1,		-- <statistic id="2872" name="Flame Leviathan kills (Ulduar 10 player)" quantity="20"/>
		[2873] = 5,		-- <statistic id="2873" name="Razorscale kills (Ulduar 10 player)" quantity="6"/>
		[2874] = 5,		-- <statistic id="2874" name="Ignis the Furnace Master kills (Ulduar 10 player)" quantity="3"/>
		[2884] = 5,		-- <statistic id="2884" name="XT-002 Deconstructor kills (Ulduar 10 player)" quantity="6"/>
		[2885] = 10,	-- <statistic id="2885" name="Assembly of Iron kills (Ulduar 10 player)" quantity="12"/>
		[2875] = 5,		-- <statistic id="2875" name="Kologarn kills (Ulduar 10 player)" quantity="18"/>
		[2882] = 5,		-- <statistic id="2882" name="Auriaya kills (Ulduar 10 player)" quantity="19"/>
		[3256] = 10,	-- <statistic id="3256" name="Hodir victories (Ulduar 10 player)" quantity="13"/>
		[3257] = 10,	-- <statistic id="3257" name="Thorim victories (Ulduar 10 player)" quantity="21"/>
		[3258] = 10,	-- <statistic id="3258" name="Freya victories (Ulduar 10 player)" quantity="15"/>
		[2879] = 10,	-- <statistic id="2879" name="Mimiron victories (Ulduar 10 player)" quantity="9"/>
		[2880] = 15,	-- <statistic id="2880" name="General Vezax kills (Ulduar 10 player)" quantity="9"/>
		[2883] = 15,	-- <statistic id="2883" name="Yogg-Saron kills (Ulduar 10 player)" quantity="8"/>
		[2881] = 10, 	-- <statistic id="2881" name="Algalon the Observer kills (Ulduar 10 player)" quantity="6"/>
		baseline = -10,
		meta = true
	},
	[L["Trial of the Crusader (10)"]] = {
		[4028] = 5,		-- <statistic id="4028" name="Victories over the Beasts of Northrend (Trial of the Crusader 10 player)" quantity="11"/>
		[4032] = 5,		-- <statistic id="4032" name="Lord Jaraxxus kills (Trial of the Crusader 10 player)" quantity="10"/>
		[4036] = 10,	-- <statistic id="4036" name="Victories over the Faction Champions (Trial of the Crusader 10 player)" quantity="9"/>
		[4040] = 10,	-- <statistic id="4040" name="Val'kyr Twins kills (Trial of the Crusader 10 player)" quantity="8"/>
		[4044] = 15,	-- <statistic id="4044" name="Times completed the Trial of the Crusader (10 player)" quantity="7"/>
		meta = true
	},
	[L["Trial of the Grand Crusader (10)"]] = {
		[4030] = 5,		-- <statistic id="4028" name="Victories over the Beasts of Northrend (Trial of the Crusader 10 player)" quantity="11"/>
		[4033] = 5,		-- <statistic id="4032" name="Lord Jaraxxus kills (Trial of the Crusader 10 player)" quantity="10"/>
		[4037] = 10,	-- <statistic id="4036" name="Victories over the Faction Champions (Trial of the Crusader 10 player)" quantity="9"/>
		[4041] = 10,	-- <statistic id="4040" name="Val'kyr Twins kills (Trial of the Crusader 10 player)" quantity="8"/>
		[4045] = 15,	-- <statistic id="4044" name="Times completed the Trial of the Crusader (10 player)" quantity="7"/>
		meta = true
	},
	[L["Trial of the Crusader (25)"]] = {
		[4031] = 5,		-- <statistic id="4028" name="Victories over the Beasts of Northrend (Trial of the Crusader 10 player)" quantity="11"/>
		[4034] = 5,		-- <statistic id="4032" name="Lord Jaraxxus kills (Trial of the Crusader 10 player)" quantity="10"/>
		[4038] = 10,	-- <statistic id="4036" name="Victories over the Faction Champions (Trial of the Crusader 10 player)" quantity="9"/>
		[4042] = 10,	-- <statistic id="4040" name="Val'kyr Twins kills (Trial of the Crusader 10 player)" quantity="8"/>
		[4046] = 15,	-- <statistic id="4044" name="Times completed the Trial of the Crusader (10 player)" quantity="7"/>
		meta = true
	},
	[L["Trial of the Grand Crusader (25)"]] = {
		[4029] = 5,		-- <statistic id="4028" name="Victories over the Beasts of Northrend (Trial of the Crusader 10 player)" quantity="11"/>
		[4035] = 5,		-- <statistic id="4032" name="Lord Jaraxxus kills (Trial of the Crusader 10 player)" quantity="10"/>
		[4039] = 10,	-- <statistic id="4036" name="Victories over the Faction Champions (Trial of the Crusader 10 player)" quantity="9"/>
		[4043] = 10,	-- <statistic id="4040" name="Val'kyr Twins kills (Trial of the Crusader 10 player)" quantity="8"/>
		[4047] = 15,	-- <statistic id="4044" name="Times completed the Trial of the Crusader (10 player)" quantity="7"/>
		meta = true
	},
}

local origSetAchievementComparisonUnit = _G.SetAchievementComparisonUnit
SetAchievementComparisonUnit = function(unit)
	inspectUnit = UnitName(unit)
	return origSetAchievementComparisonUnit(unit)
end

local function buildLookupWeights(set)
	for dungeon, stats in pairs(set) do
		if type(stats) == "table" then
			buildLookupWeights(stats)
		elseif tonumber(dungeon) and tonumber(dungeon) > 0 then
			weightLookup[dungeon] = stats
		end
	end
end

function mod:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("ExperienceDB", defaults)
	db = self.db.realm
	profile = self.db.profile
	if db.cache then db.cache = nil end
	if profile.cache then profile.cache = nil end
	
	buildLookupWeights(dungeons)
	for k, v in pairs(dungeons) do
		tinsert(dungeonKeys, k)
	end
	table.sort(dungeonKeys)
	
	LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable("Experience", options)	
	configFrame = ACD3:AddToBlizOptions("Experience")
	
	for k, v in pairs(dungeons) do
		options.args.dungeons.args[k:gsub(" ", "_")] = {
			type = "toggle",
			name = k,
			get = function()
				return profile.excludeDungeons[k] ~= true
			end,
			set = function(info, v)
				profile.excludeDungeons[k] = not v and true or nil
			end
		}
	end
	
	SLASH_EXPERIENCE1, SLASH_EXPERIENCE2 = '/xp', '/experience';
	SlashCmdList["EXPERIENCE"] = mod.Toggle
end

function mod:OnEnable()
	self:RegisterEvent("INSPECT_ACHIEVEMENT_READY")
	self:RegisterEvent("PLAYER_TARGET_CHANGED")
end

function mod:Toggle()
	if mod:IsEnabled() then
		mod:Disable()
		print("|cff9999ffExperience|r is now |cffff0000disabled|r.")
		ldbObj.text = "|cffff0000Disabled|r"
	else
		mod:Enable()
		print("|cff9999ffExperience|r is now |cff00ff00enabled|r.")
		ldbObj.text = "|cff00ff00Enabled|r"
	end
end

local function getStat(name, stat)
	cache[name] = cache[name] or {}
	local t = cache[name]
	t.time = GetTime()
	t[stat] = tonumber(GetComparisonStatistic(stat)) or 0
end

local function getAchievmentSetScores(name, set)
	if not name then return false end
	
	for dungeon, stats in pairs(set) do
		if type(stats) == "table" then
			getAchievmentSetScores(name, stats)
		else
			if tonumber(dungeon) and tonumber(dungeon) > 0 then
				getStat(name, dungeon)
			end
		end
	end
	return true
end

local function prepareTooltip(unit)
	local tip = tooltip
	if tip == tooltip then
		tip:ClearLines()
		tip:Show()
		tip:SetOwner(UIParent, "ANCHOR_PRESERVE")
		tip:AddLine(("|cff5566ff%s:|r %s"):format(L["XP"], unit))
	end
	return tip
end

function mod:INSPECT_ACHIEVEMENT_READY(...)
	if inspectUnit and getAchievmentSetScores(inspectUnit, dungeons) then
		self:UpdateTooltip(inspectUnit)
		inspectUnit = nil
	end
end

function mod:UpdateTooltip(unit)
	if not unit or unit ~= currentUnit then return end
	
	local grandScore, grandTotal = 0, 0
	local tip = prepareTooltip(unit)
	for _, dungeon in ipairs(dungeonKeys) do
		local set = dungeons[dungeon]
		local data = set.meta and set or metas[dungeon]
		if data then
			local score, total = self:GetScore(unit, data)
			if score and total and not profile.excludeDungeons[dungeon] then
				if score > 0 or not profile.hideEmpty then
					local pct =  score / total
					local r, g, b = 0.4, 0.6, 1
					if pct <= 1 then
						r = pct <= 0.5 and 1 or (1 - pct)
						g = pct >= 0.5 and 1 or pct
						b = 0
					end
					local right
					if profile.showDetail then
						right = ("%2.1f%% [%d/%d]"):format(pct * 100.0, score, total)
					else
						right = ("%2.1f%%"):format(pct * 100.0)
					end
					tip:AddDoubleLine(
						dungeon, 
						right,
						1, 1, 1,
						r, g, b
					)
				end				
				grandScore = grandScore + score
				grandTotal = grandTotal + total
			end
		end
	end	
	local pct = math.min(1, grandScore / grandTotal)
	local r = pct <= 0.5 and 1 or (1 - pct)
	local g = pct >= 0.5 and 1 or pct
	tip:AddDoubleLine(
		L["Total Experience"], 
		("%2.1f%% (%2.0f/%2.0f)"):format(pct * 100.0, grandScore, grandTotal),
		1, 1, 1,
		r, g, 0
	)
	tip:Show()
end

function mod:PLAYER_TARGET_CHANGED()
	self:Inspect("target")
end

function mod:ValidToInspect()
	if UnitAffectingCombat("player") then return false end
	return true
end

function mod:Inspect(unit)
	if not self:ValidToInspect() then return end
	
	if UnitName(unit) and UnitLevel(unit) == 80 then
		currentUnit = UnitName(unit)
		if cache[currentUnit] and GetTime() - cache[currentUnit].time < 3600 then
			self:UpdateTooltip(currentUnit)
			return
		else
			ClearAchievementComparisonUnit()
			SetAchievementComparisonUnit(unit)
		end
	end
end

local function getScore(unit, id)
	return math.min(cache[unit][id] or 0, profile.experiencedKills) * weightLookup[id], profile.experiencedKills * weightLookup[id]
end

function mod:GetScore(unit, set)
	if not cache[unit] then return end
	local score, total = 0, 0
	if #set > 0 then		
		for _, k in ipairs(set) do
			local ps, pt = getScore(unit, k)
			score = score + ps
			total = total + pt
		end
	else
		for k, v in pairs(set) do
			if tonumber(k) and tonumber(k) > 0 then
				local ps, pt = getScore(unit, k)
				score = score + ps
				total = total + pt
			end
		end
	end
	total = total + ((set.baseline or 0) * profile.experiencedKills)
	return score / profile.experiencedKills, total / profile.experiencedKills
end
